{{      
************************************************
* Propeller SPI Engine  ... Spin Version  v1.1 *
* Author: Beau Schwabe                         *
* Copyright (c) 2009 Parallax                  *
* See end of file for terms of use.            *
************************************************

Revision History:
  V1.0  - original program
  V1.1  - spiff up for Gold Standard by Perry Harrington

This SPIN object module is a simple software interface for doing clocked serial
communications.  There are several types of clocked serial communications, this
module is named for the SPI clocked serial protocol.

▶MSB vs LSB◀
When sending or receiving serial data, you can send it using a matrix of options.
The first consideration is the endianess of the target device, does it want to
speak in Little endian (LSB first) or Big endian (MSB first).

To visualize endianness, Big endian sends the high bit first (eg: bit 31), Little
endian sends the low bit first (eg: bit 0).  High being the leftmost, low being
the right most.

When sending data this option is specified as MSB or LSB as the mode
parameter to the SHIFTOUT routine.

When reading data from a device, you specify one of MSB or LSB and add PRE or POST
to indicate PRE clock or POST clock.

▶Pre-clock vs Post-clock◀
The second consideration is pre clocking or post clocking of incoming data.  This
determines when the data is valid when received from the device.  If you look at a
timing diagram, you will see the clock cycling before the data, or the clock
cycling after the data is presented to the bus.

This is visualized thusly:

Pre clock:

Initial High state
─────────
──0123───
   R X R X R X R X

Initial Low state
─────
──0123───
   R X R X R X R X

Post clock:

Initial High state
─────────
────0123─
     R X R X R X R X

Initial Low state
──────
────0123─
     R X R X R X R X

The difference with Pre clocking and Post clocking is that the data is available
either before the clock, or after the clock.  Referring to the diagrams above, the
SHIFTIN routine reads the data line where the R is.

Pre-clocking or Post-clocking applies only to the mode parameter of the SHIFTIN
routine.  This is specified with one of the PRE and POST constants as the mode
parameter.  For MSB first and pre clocking you would give the parameter MSB + PRE and
for LSB and post clocking you would use LSB + POST

IMPORTANT NOTE:

The initial clock state is important as to when the device makes the data available.
A device can make the data available when the clock is high or low.  This means the
clock will be going low if the initial state is high or the clock will be going high
if the initial state is low.

If the initial clock state is not set correctly, the data will be read out of sync
and the initial bit read will either be lost or a bit added.  You should validate
your device output carefully, even trying both initial states to see which produces
valid data.

▶Initial clock state◀

When communicating with serial devices, they will indicate whether a clock is valid
low or valid high.  Typically a low active clock is indicated by an inverting bar or
slash in the name of the clock signal, a high active clock is generally indicated by
the lack of an inverting bar or slash.  You must specify the inactive state of the clock,
for an active low clock you specify HIGH as the inital state, for active high clock
you specify LOW as the initial state.

}}
CON

  MSB = 0       'MSB first
  LSB = 1       'LSB first
  PRE = 0       'Pre clocked input
  POST = 2      'Post clocked input
  HIGH = 1      'Inactive state is high
  LOW = 0       'Inactive state is low

{ The constants below are for backwards compatibility, the constants above are a
  simplified version of the same. }

  #0,MSBPRE,LSBPRE,MSBPOST,LSBPOST                    '' Used for SHIFTIN routines
''
'' MSBPRE   - Most Significant Bit first ; data is valid before the clock
'' LSBPRE   - Least Significant Bit first ; data is valid before the clock
'' MSBPOST  - Most Significant Bit first ; data is valid after the clock
'' LSBPOST  - Least Significant Bit first ; data is valid after the clock


  #0,MSBFIRST,LSBFIRST                                '' Used for SHIFTOUT routines
''
'' LSBFIRST - Least Significant Bit first ; data is valid after the clock
'' MSBFIRST - Most Significant Bit first ; data is valid after the clock

VAR
  long          clockDelay, clockState

PUB Dummy
'This is not a top level object and must be called using the OBJ interface

PUB Start(_clockDelay, _clockState)
{{
  Initialize the clock rate and inactive state of the clock.

  parameters:
  ClockDelay                    long                    clock rate in 10µs units
  ClockState                    long                    inactive state of clock line, one of #HIGH or #LOW
  return:                       none

  example usage:    SPI.Start(400,HIGH) 'active low 400Khz signalling rate
  expected results: ClockState is initialized and ClockDelay is computed
}}

  clockState := _clockState                             'inactive state of clock line

{ Set signalling rate delay and ensure delay minimum value isn't less than the execution
  time of the waitct SPIN routine, see Propeller manual pg 219 (WAITCNT).  Also you will
  note the delay is reduced by the number of clocks it takes to run the clocking
  routine, this was determined by profiling the function.
}
  clockDelay := ((clkfreq / 100000 * _clockDelay) - 4296) #> 381
        
PUB ShiftOut(dPin, cPin, mode, bits, value)
{{
  Shift data out the selected pin, using the endian specified.

  parameters:
  dPin    long                  Data pin to transmit on, range 0-31
  cPin    long                  Clock pin to signal on, range 0-31
  mode    long                  The endian bit order, one of #MSB or #LSB
  bits    long                  The number of bits to transmit, range 0-31
  Value   long                  The data value to transmit
  return: none

  example usage: SPI.SHIFTOUT(0,1,SPI#MSB,8,$AA)
  expected results:  Bits of Value are transmitted on Dpin signalled with Cpin
}}

    dira[dPin]~~                                        ' make Data pin output
    outa[cPin] := clockState                            ' set initial clock state
    dira[cPin]~~                                        ' make Clock pin output

    if mode == LSB                                      ' LSBFIRST
       value <-= 1                                      ' pre-align lsb
       repeat bits                                      ' send Bits of Value
         outa[dPin] := (value ->= 1) & 1                ' output data bit
         PostClock(cPin)                                ' toggle clock pin

    if mode == MSB                                      ' MSBFIRST
       value <<= (32 - bits)                            ' pre-align msb
       repeat bits                                      ' send Bits of Value
         outa[dPin] := (value <-= 1) & 1                ' output data bit
         PostClock(cPin)                                ' toggle clock pin

PUB ShiftOut(dPin, cPin, mode, bits)
{{
  Shift data in from Dpin using endian and pre-clock or post-clocking

  parameters:
  dPin    long                  Data pin to receive on, range 0-31
  cPin    long                  Clock pin to signal on, range 0-31
  mode    long                  Endianness and Clock mode, #MSB or #LSB + #PRE or #POST
  bits    long                  Number of bits to receive

  return: result, data received on Dpin

  example usage: value := SPI.SHIFTIN(0,1,SPI#MSB + SPI#POST, 8)
  expected results: Bits of data read on Dpin with MSB, LSB and PRE,POST settings
}}

    dira[dPin]~                                         ' make dpin input
    outa[cPin] := clockState                            ' set initial clock state
    dira[cPin]~~                                        ' make cpin output

    if mode == MSBPRE                                   ' MSBPRE
       repeat bits
         result := (result << 1) | ina[dPin]            ' read Dpin, rotate into result
         PostClock(cPin)                                ' clock Cpin
          
    if mode == LSBPRE                                   ' LSBPRE
       repeat bits +1
         result := (result >> 1) | (ina[dPin] << 31)    ' read Dpin, rotate into result
         PostClock(cPin)                                ' clock Cpin
       result >>= (32 - bits)                           ' zero extend unused bits

    if mode == MSBPOST                                  ' MSBPOST
       repeat bits
         PreClock(cPin)                                 ' clock Cpin
         result := (result << 1) | ina[dPin]            ' read Dpin, rotate into result
 
    if mode == LSBPOST                                  ' LSBPOST
      repeat bits + 1
        PreClock(cPin)                                  ' clock Cpin
        result := (result >> 1) | (ina[dPin] << 31)     ' read Dpin, rotate into result
      result >>= (32 - bits)                            ' zero extend unused bits

  'result is an alias for the return value, we don't need to explicitly add a return statement

PRI PostClock(_Cpin)
{{
  Do post data clocking and signalling delay

  parameters:
  _Cpin   long                  Clock pin

  return: none

  example usage: this is private function, see above
}}

    waitcnt(cnt+clockDelay)
    !outa[_Cpin]
    waitcnt(cnt+clockDelay)
    !outa[_Cpin]

PRI PreClock(_Cpin)
{{
  Do pre data clocking and signalling delay

  parameters:
  _Cpin   long                  Clock pin

  return: none

  example usage: this is a private function, see above
}}

    !outa[_Cpin]
    waitcnt(cnt+clockDelay)
    !outa[_Cpin]
    waitcnt(cnt+clockDelay)

DAT
{{
┌──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┐
│                                                   TERMS OF USE: MIT License                                                  │                                                            
├──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┤
│Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    │ 
│files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    │
│modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software│
│is furnished to do so, subject to the following conditions:                                                                   │
│                                                                                                                              │
│The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.│
│                                                                                                                              │
│THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          │
│WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         │
│COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   │
│ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         │
└──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┘
}}
